package net.knarcraft.permissionsigns.listener;

import net.knarcraft.permissionsigns.PermissionSigns;
import net.knarcraft.permissionsigns.container.PermissionSign;
import net.knarcraft.permissionsigns.container.SignCreationRequest;
import net.knarcraft.permissionsigns.formatting.StringFormatter;
import net.knarcraft.permissionsigns.formatting.TranslatableMessage;
import net.knarcraft.permissionsigns.formatting.Translator;
import net.knarcraft.permissionsigns.manager.EconomyManager;
import net.knarcraft.permissionsigns.manager.PermissionManager;
import net.knarcraft.permissionsigns.manager.SignManager;
import org.bukkit.Material;
import org.bukkit.Tag;
import org.bukkit.block.Block;
import org.bukkit.block.Sign;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.player.PlayerInteractEvent;

import java.util.Arrays;
import java.util.StringJoiner;

/**
 * A listener for all events related to signs
 */
public class SignListener implements Listener {

    /**
     * This event handler detects if a player clicks a button or a sign
     *
     * @param event <p>The player interact event which was triggered</p>
     */
    @EventHandler
    public void onPlayerInteract(PlayerInteractEvent event) {
        Player player = event.getPlayer();
        Block block = event.getClickedBlock();

        if (block == null) {
            return;
        }

        Material material = block.getBlockData().getMaterial();
        if (!Tag.SIGNS.isTagged(material) && !Tag.WALL_SIGNS.isTagged(material)) {
            return;
        }

        Sign sign = (Sign) block.getState();
        handleSignClick(sign, player);
    }

    /**
     * Handles the right-clicking action
     *
     * @param sign   <p>The clicked sign</p>
     * @param player <p>The player that clicked the sign</p>
     */
    private void handleSignClick(Sign sign, Player player) {
        //Check if the sign is a registered permission sign
        PermissionSign permissionSign = SignManager.getSign(sign.getLocation());
        if (permissionSign != null) {
            if (!player.hasPermission("permissionsigns.use")) {
                player.sendMessage(StringFormatter.getTranslatedErrorMessage(TranslatableMessage.INTERACT_PERMISSION_MISSING));
                return;
            }
            handlePermissionSignInteract(permissionSign, player);
            return;
        }
        //TODO: Display information about granted permissions when shift-clicking?

        //Check if the player has a creation request that can be fulfilled
        SignCreationRequest request = PermissionSigns.getSignCreationRequest(player.getUniqueId());
        if (request != null) {
            registerPermissionSign(sign, request, player);
        }
    }

    /**
     * Handles the interaction with a permission sign
     *
     * @param permissionSign <p>The permission sign that was interacted with</p>
     * @param player         <p>The player that interacted with the permission sign</p>
     */
    private void handlePermissionSignInteract(PermissionSign permissionSign, Player player) {
        //Don't allow players to pay for permissions they already own
        boolean hasAllPermissions = true;
        for (String permissionNode : permissionSign.getPermissionNodes()) {
            if (permissionNode.contains(":")) {
                String[] permissionParts = permissionNode.split(":");
                if (!PermissionManager.hasPermission(player, permissionParts[1], permissionParts[0])) {
                    hasAllPermissions = false;
                }
            } else if (!player.hasPermission(permissionNode)) {
                hasAllPermissions = false;
            }
        }
        if (hasAllPermissions) {
            player.sendMessage(StringFormatter.getTranslatedErrorMessage(TranslatableMessage.ALREADY_HAS_PERMISSIONS));
            return;
        }

        if (EconomyManager.canAfford(player, permissionSign.getCost())) {
            performPermissionTransaction(player, permissionSign);
        } else {
            player.sendMessage(StringFormatter.getTranslatedErrorMessage(TranslatableMessage.CANNOT_AFFORD));
        }
    }

    /**
     * Performs the permission transaction taking payment and granting permissions
     *
     * @param player         <p>The player to give permissions to</p>
     * @param permissionSign <p>The permission sign the player clicked</p>
     */
    private void performPermissionTransaction(Player player, PermissionSign permissionSign) {
        EconomyManager.withdraw(player, permissionSign.getCost());
        StringJoiner permissionsJoiner = new StringJoiner(", ");
        for (String permissionNode : permissionSign.getPermissionNodes()) {
            //Only grant the permission if not already owned
            boolean hasPermission;
            if (permissionNode.contains(":")) {
                String[] permissionParts = permissionNode.split(":");
                hasPermission = PermissionManager.hasPermission(player, permissionParts[1], permissionParts[0]);
            } else {
                hasPermission = player.hasPermission(permissionNode);
            }
            if (!hasPermission) {
                permissionsJoiner.add(permissionNode);
                if (permissionSign.getDuration() == 0) {
                    PermissionManager.addPermission(player, permissionNode);
                } else {
                    PermissionManager.addTemporaryPermission(player, permissionNode, permissionSign.getDuration());
                }
            }
        }

        String grantedPermissions = permissionsJoiner.toString();
        if (permissionSign.getDuration() > 0) {
            String successMessage = StringFormatter.getTranslatedInfoMessage(TranslatableMessage.PERMISSIONS_GRANTED);
            String timeUnit = Translator.getTranslatedMessage(TranslatableMessage.SIGN_TIME_UNIT);
            player.sendMessage(StringFormatter.replacePlaceholders(successMessage, new String[]{"{permissions}", "{time}"},
                    new String[]{grantedPermissions, permissionSign.getDuration() + " " + timeUnit}));
        } else {
            player.sendMessage(StringFormatter.replacePlaceholder(StringFormatter.getTranslatedInfoMessage(
                    TranslatableMessage.PERMISSIONS_GRANTED_PERMANENTLY), "{permissions}", grantedPermissions));
        }
    }

    /**
     * Registers a new permission sign if the given sign is empty
     *
     * @param sign    <p>The sign to register as a permission sign</p>
     * @param request <p>The creation request to fulfill</p>
     * @param player  <p>The player that interacted with the sign</p>
     */
    private void registerPermissionSign(Sign sign, SignCreationRequest request, Player player) {
        String[] lines = sign.getLines();
        //Don't allow non-empty signs to be overwritten
        if (!Arrays.stream(lines).allMatch(String::isEmpty)) {
            return;
        }

        //Register the sign and remove the request
        PermissionSign permissionSign = request.getPermissionSign();
        permissionSign.setSignLocation(sign.getLocation());
        SignManager.addSign(permissionSign);
        PermissionSigns.cancelSignCreationRequest(player.getUniqueId());

        //Display the permission sign information on the sign
        String[] signLines = permissionSign.getSignLines();
        for (int i = 0; i < signLines.length; i++) {
            sign.setLine(i, signLines[i]);
        }
        sign.update();
        player.sendMessage(StringFormatter.getTranslatedInfoMessage(TranslatableMessage.PERMISSION_SIGN_CREATED));
    }

}
